unit UdpClass;

interface

uses Windows, Messages, SysUtils, Variants, Classes, Dialogs, Forms,
 WSocket, WSockets, ExtCtrls, POSITIONS;

type

// ------------ Mise en forme de la classe -------------------------------------
TUdpClient = class(TComponent)

  constructor Create(AOwner:Tcomponent;P:string;A:string;S:integer;E:TFrmPosEcran);
  Destructor Destroy;

  private
    // -- Les Variables --
    Port:string;
    Adress:string;
    UDPSocket: TWSocket;

    Timer:TTimer;
    Interval:Cardinal;

    Ecran:TFrmPosEcran;

    BufferTMot : array [0..5] of string;
    BufferTMotTeller : Integer;

    NBEssais:integer;   {Compteur de tentatives d'appels a la methode declenche par le timer}
    UdpClientOccuper : boolean;   {Si le socket est occup}
    UdpClientFoB : boolean;       {Ce qu'il doit faire,False -> FW, True -> Butees}

    Data:string;                  {Ce qu'il reoit}

    FW1:string;                   {Les Flagwords, qu indique quel moteur tourne}
    FW2:string;                   { quelle direction}
    FW3:string;
    FW4:string;
    FW5:string;

    MEW0:string;                  {Input word0:consiste quelques valeurs butes}
    MMW50:string;                 {Flag word50:consiste quelques valeurs butes}

    SV,EV,CV,PV,PIV:integer;      {Valeurs des bute, chacun rprsente un moteur}

    CoudePosition,PoignetPosition:integer;

    // -- Les Methodes --
    procedure GetFlagWords(Sender:TObject);
    procedure GetButees(Sender:TObject);
    procedure GetValues();
    procedure UDPSocketDataAvailable(Sender: TObject; Error: Word);
    procedure AnalyseData();
    function BinToDig(BinStr:String):Integer;
    function CalculeMew0EnHex():string;
    function pow(base, power: integer): integer;
    procedure SendButee();
    procedure UDPConnect;

  public

    procedure CommenceComIPCSR(); //Commence la com. avec l'IPC sans robot
    procedure CommenceComIPCAR(); //Commence la com. avec l'IPC avec robot

    procedure ArreteComIPC();

    procedure SetButee(RV:String);
    // -- Properties --
    property PFW1 : string read FW1; //FW1: fonctionnement de pince
    property PFW2 : string read FW2; //FW2: fonctionnement de poignet
    property PFW3 : string read FW3; //FW3: fonctionnement de coude
    property PFW4 : string read FW4; //FW4: fonctionnement de epaule
    property PFW5 : string read FW5; //FW5: fonctionnement de socle

    property PSV : integer read SV;
    property PEV : integer read EV;
    property PCV : integer read CV;
    property PPV : integer read PV;
    property PPIV : integer read PIV;

    property PCoudePosition : integer read CoudePosition;
    property PPoignetPosition : integer read PoignetPosition;

end;
// ------------ Fin de la mise en forme ----------------------------------------

implementation
uses unit1;

// ------------ Implementation des methodes de TUDPCLIENT ----------------------

constructor TUdpClient.Create(AOwner:Tcomponent;P:string;A:string;S:integer;E:TFrmPosEcran);
begin
  Port := P;
  Adress := A;
  UDPSocket := TWSocket.Create(Self);
  UDPSocket.OnDataAvailable := UDPSocketDataAvailable;

  Interval := S;
  Timer := TTimer.Create(Self);
  Timer.Interval := Interval;
  Timer.Enabled := False;

  Ecran := E;
  BufferTMotTeller := 1;
  UdpClientOccuper := false;

  FW1 := '193';
  FW2 := '194';
  FW3 := '195';
  FW4 := '196';
  FW5 := '197';

  MEW0 := '10010000';    // a l'initialisation: Socle  droite, Epaule en haut
  MMW50 := '10101000';   //                     poignet  droite, pince ferme, coude en haut

  SV:=2; EV:=2; CV:=2; PV:=2; PIV:=2;

  NBEssais:=0;

end;

Destructor TUdpClient.Destroy;
begin
    Timer.Free;
    UDPSocket.Free;
    inherited Destroy;
end;

// ------------ Procedures Priv------------------------------------------------

{Demande du service UDP
appel de la mthode connect si pas deja connecte}
Procedure TUdpClient.UDPConnect;
begin
if UDPSocket.state = wsconnected then
  UDPSocket.close;
UdpSocket.Proto:='udp';
UdpSocket.Addr:=Adress;
UdpSocket.Port:=Port;
UdpSocket.Connect;
end;

{Met tous les mots qu'on a besoin pour retrouver les flagwords dans un chane
 et commence la boucle des envoyer.}
procedure TUdpClient.GetFlagWords(Sender:TObject);
begin
  if UdpClientOccuper = False then
  begin
    NBEssais:=0;
    BufferTmotTeller := 1;
    BufferTMot[1] := 'DMW1'; //Pince
    BufferTMot[2] := 'DMW2'; //Poignet
    BufferTMot[3] := 'DMW3'; //Coude
    BufferTMot[4] := 'DMW4'; //Epaule
    BufferTMot[5] := 'DMW5'; //Socle
    UdpClientOccuper := True;
    GetValues();
  end
  else
  begin
    inc (NBEssais);
    if NBEssais>10 then
      UdpClientOccuper:=false;  //Si on perd une trame, on redemarre le debut de l'envoi
  end;
end;

{Met tous les mots qu'on a besoin pour retrouver les valeurs des butes et des
 angles de la coude et du poignet. Commence la boucle des envoyer.}
procedure TUdpClient.GetButees(Sender:TObject);
begin
  if UdpClientOccuper = False then
  begin
    NBEssais:=0;
    BufferTmotTeller := 1;
    BufferTMot[1] := 'DEW0'; //Epaule et Socle
    BufferTMot[2] := 'DMW50'; //Pince et poignet
    BufferTMot[3] := 'DEW12'; //Angle coude
    BufferTMot[4] := 'DEW13'; //Angle poignet
    UdpClientOccuper := True;
    GetValues();
  end
  else
  begin
    inc (NBEssais);
    if NBEssais>10 then
      UdpClientOccuper:=false;  //Si on perd une trame, on redemarre le debut de l'envoi
  end;
end;

{Envoyer un mot au IPC, il peut savoir quel mot avec le compteur BufferTMotTeller
 Qui est augmenter aprs chaque reu.}
procedure TUdpClient.GetValues();
begin

//  UdpSocket.Proto:='udp';
//  UdpSocket.Addr:=Adress;
//  UdpSocket.Port:=Port;
//  UdpSocket.Connect;
  UDPConnect;
  UdpSocket.SendStr(BufferTMot[BufferTMotTeller]);

end;

{Chaque fois il a envoi un mot, il attande sur une rponse. Il n'envoie pas un
 nouveau mot parce que le variable UdpClientOccuper est Vrai pour le moment. Si
 la rponse arrive, il l'analyse et augmente le compteur}
procedure TUdpClient.UDPSocketDataAvailable(Sender: TObject; Error: Word);
begin

  Data := UdpSocket.ReceiveStr;
  FrmSimulRobot.mmoServerMessages.Lines.Add(inttostr(length(data))+':'+data);
  UdpSocket.Close;
  if length(data)>5 then

    begin
    AnalyseData();

  if UdpClientFoB = False then //recevoir les FlagWords
  begin
    if BufferTMotTeller < 5 then
    begin
      BufferTMotTeller := BufferTMotTeller + 1;
      GetValues();
    end
    else
    begin
      SendButee();
    end;
  end;

  if UdpClientFoB = True then //recevoir les Butees
  begin
    if BufferTMotTeller < 4 then
    begin
      BufferTMotTeller := BufferTMotTeller + 1;
      GetValues();
    end;
  end;
  end;
end;

{Une rponse consiste de deux partie, la premire avec le mot envoi et la
 deuxime avec la vrai rponse. Cela on doit d'abord driver. Aprs on peut
 voir avec la premire partie qoui on doit faire avec la deuxime.}
procedure TUdpClient.AnalyseData();
var
  reqmot,valeur : String;
  t1,t2,chiffre : Integer;
begin

  //cherche dans le resultat le nom du demande
  reqmot := '';
  for t1 := 1 to 3 do
  begin
    reqmot := reqmot + Data[t1];
  end;

  //cherche dan le resultat le valeur du demande
  t1 := pos('=',Data);
  valeur := '';
  for t2 := t1+1 to length(Data) do
  begin
    valeur := valeur + Data[t2];
  end;

  if UdpClientFoB = False then //recevoir les FlagWords
  begin
    if reqmot = 'DMW' then
    begin
      case Data[4] of
        '1':FW1 := valeur;
        '2':FW2 := valeur;
        '3':FW3 := valeur;
        '4':FW4 := valeur;
        '5':FW5 := valeur;
      end;
    end;
  end;

  if UdpClientFoB = True then //recevoir les valeurs des angles et des butes
  begin

    chiffre := StrToInt(valeur);

    if (reqmot = 'DEW') And (BufferTMotTeller = 1) then
    begin                                                   {Mettre un masque}
      SV:=2; EV:=2;
      if (chiffre And $08) = $08 then //Butee Epaule basse
        EV := 3;
      if (chiffre And $10) = $10 then //Butee Epaule haute
        EV := 1;
      if (chiffre And $20) = $20 then //Butee Socle gauche
        SV := 1;
      if (chiffre And $40) = $40 then //Butee Socle milieux
        SV := 2;
      if (chiffre And $80) = $80 then //Butee Socle droite
        SV := 3;
    end;
    if (reqmot = 'DMW') And (BufferTMotTeller = 2) then
    begin
      PV:=2; PIV:=2;
      if (chiffre And $10) = $10 then //Butee Pince fermee
        PIV := 1;
      if (chiffre And $20) = $20 then //Butee Pince ouvert
        PIV := 3;
      if (chiffre And $40) = $40 then //Butee Poignet gauche
        PV := 1;
      if (chiffre And $80) = $80 then //Butee Poignet droite
        PV := 3;
    end;
    if (reqmot = 'DEW') And (BufferTMotTeller = 3) then //Angle Coude
    begin
      chiffre := -((chiffre div 9)-237);
      CoudePosition := chiffre;
      Ecran.MiseDegreeCoude(chiffre);
    end;
    if (reqmot = 'DEW') And (BufferTMotTeller = 4) then //Angle Poignet
    begin
      PoignetPosition := chiffre;
      Ecran.MiseDegreePoignet(chiffre);
      UdpClientOccuper := False;
    end;

    ecran.MiseLesValeurs(SV,EV,CV,PV,PIV);

  end;
end;

{Si un des moteurs est venu en bute avec la simulation des Flag Words on doit
 crire l'crire dans le mot input et le mot flag word.}
procedure TUdpClient.SendButee();
var
  SS:String;
begin
//  UdpSocket.Proto:='udp';
//  UdpSocket.Addr:=Adress;
//  UdpSocket.Port:=Port;
//  UdpSocket.Connect;
  UDPConnect;
  SS := 'MEW0=' + IntToStr(BinToDig(MEW0));
  UdpSocket.SendStr(SS);
  UdpSocket.Close;

  sleep(20);

//  UdpSocket.Proto:='udp';
//  UdpSocket.Addr:=Adress;
//  UdpSocket.Port:=Port;
//  UdpSocket.Connect;
  UDPConnect;

  SS := 'MMW50=' + IntToStr(BinToDig(MMW50));
  UdpSocket.SendStr(SS);
  UdpSocket.Close;

  UdpClientOccuper := False;
end;

{Calculation d'un numro binaire en hexadecimal}
function TUdpClient.CalculeMew0EnHex():String;
const
  BinArray: array[0..15, 0..1] of string =
    (('0000', '0'), ('0001', '1'), ('0010', '2'), ('0011', '3'),
     ('0100', '4'), ('0101', '5'), ('0110', '6'), ('0111', '7'),
     ('1000', '8'), ('1001', '9'), ('1010', 'A'), ('1011', 'B'),
     ('1100', 'C'), ('1101', 'D'), ('1110', 'E'), ('1111', 'F'));
var
  Error: Boolean;
  j: Integer;
  BinPart,BinStr: string;
begin
  BinStr := MEW0;
  Result:='';

  Error:=False;
  for j:=1 to Length(BinStr) do
    if not (BinStr[j] in ['0', '1']) then
    begin
      Error:=True;
      ShowMessage('This is not binary number');
      Break;
    end;

  if not Error then
  begin
    case Length(BinStr) mod 4 of
      1: BinStr:='000'+BinStr;
      2: BinStr:='00'+BinStr;
      3: BinStr:='0'+BinStr;
    end;

    while Length(BinStr)>0 do
    begin
      BinPart:=Copy(BinStr, Length(BinStr)-3, 4);
      Delete(BinStr, Length(BinStr)-3, 4);
      for j:=1 to 16 do
        if BinPart=BinArray[j-1, 0] then
          Result:=BinArray[j-1, 1]+Result;
    end;
  end;
end;

{Calculation d'un numro binaire en decimal}
function TUdpClient.BinToDig(BinStr:String):Integer;
var
  i,counter: Integer;
begin
  if length(BinStr)>16 then
    raise ERangeError.Create(#13+BinStr+#13+
            'is not within the valid range of a 16 bit binary.'+#13);

  Result:=0;

  for counter:=1 to length(BinStr) do
      if BinStr[Counter]='1' then
        Result:=Result+pow(2,length(BinStr)-counter);

end;

{Elevation  une puissance}
function TUdpClient.pow(base, power: integer): integer;
var counter : integer;
begin
  Result:=1;

  for counter:=1 to power do
    Result:=Result*base;
end;


// ------------ Procedures Publique---------------------------------------------
{Commence la communication avec l'IPC sans Robot}
procedure TUdpClient.CommenceComIPCSR();
begin

//Activation de la lecture des commandes de moteur
  Timer.OnTimer := GetFlagWords;
  UdpClientFoB := False;
  Timer.Enabled := True;
end;

{Commence la communication acev l'IPC avec Robot}
procedure TUdpClient.CommenceComIPCAR();
begin

//Activation de la lecture des positions du robot
  Timer.OnTimer := GetButees;
  UdpClientFoB := True;
  Timer.Enabled := True;
end;

{Arrter toutes les communications}
procedure TUdpClient.ArreteComIPC();
begin
  Timer.Enabled := False;
 
end;

{Mettre un valeur d'un bute d'un de les moteurs}
//MEW0 est une chaine de 8 caracteres '0' ou '1', numrote de 1  8
//     MEW0= 1 2 3 4 5 6 7 8      correspond 
//     IW0 = 7 6 5 4 3 2 1 0
// de mme pour FW50              correspond 
//     MMW50= 1 2 3 4 5 6 7 8
//     FW50 = 7 6 5 4 3 2 1 0
procedure TUdpClient.SetButee(RV:String);
begin
  if RV = '11' then //butee socle gauche
  begin
    MEW0[3] := '1'; //I0.5
    MEW0[1] := '0'; // a cause de l'init
    end;
  if RV = '13' then //butee socle droite
  begin
    MEW0[1] := '1'; //I0.7
    MEW0[3] := '0'; // a cause de l'init
  end;  
  if RV = '12' then //Socle ne pas butee
  begin
    MEW0[3]:='0';
    MEW0[1]:='0';
  end;
  if RV = '21' then // Epaule haute
  begin
    MEW0[4] := '1'; // I0.4
    MEW0[5] := '0';
    end;
  if RV = '23' then // Epaule basse
  begin
    MEW0[5] := '1'; // I0.3
    MEW0[4] :='0';
    end;
  if RV = '22' then // Epaule ne pas butee
  begin
    MEW0[4] := '0';
    MEW0[5] := '0';
  end;
  if RV = '41' then //butee Poignet gauche
  begin
    MMW50[2] := '1';//F50.6
    MMW50[1] := '0';
  end;
  if RV = '43' then //butee Poignet droite
  begin
    MMW50[1] := '1';//F50.7
    MMW50[2] := '0';
    end;
  if RV = '42' then //Poignet ne pas butee
  begin
    MMW50[1] := '0';
    MMW50[2] := '0';
  end;
  if RV = '51' then // butee pince ferme
  begin
    MMW50[3] := '1';// F50.4
    MMW50[4] := '0';
  end;
  if RV = '53' then // Butee pince ouverte
  begin
    MMW50[4] := '1';//F50.5
    MMW50[3] := '0';
    end;
  if RV = '52' then
  begin
    MMW50[3] := '0';
    MMW50[4] := '0';
  end;
  if RV = '31' then // butee coude haut
  begin
    MMW50[5] := '1';// F50.3
    MMW50[6] := '0';
    end;
  if RV = '33' then // Butee coude bas
  begin
    MMW50[6] := '1';//F50.2
    MMW50[5] :='0';
    end;
  if RV = '32' then
  begin
    MMW50[5] := '0';
    MMW50[6] := '0';
  end;
end;

end.
